package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.AdvertisementBasedRspecGeneratorConfig;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfig;
import be.iminds.ilabt.jfed.testing.base.ApiTestConfigProperty;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStepConfig;
import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.jetbrains.annotations.Contract;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.*;
import java.util.function.Supplier;

@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public class GuiLogicTestConfig implements ApiTestConfig {
    private static final Logger LOG = LoggerFactory.getLogger(GuiLogicTestConfig.class);

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy;
    @Nonnull private final String amVersion;
    @Nonnull private final SSHKeys sshKeys;
    @Nonnull private final List<Resource> resources;
    @Nonnull private final List<AnsibleTest> ansibleTests;
    @Nonnull private final NodeLoginTest nodeLoginTest;

    public GuiLogicTestConfig() {
        this.amConnectionProxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
        this.amVersion = "auto";
        this.sshKeys = new SSHKeys();
        this.resources = Collections.singletonList(new Resource());
        this.ansibleTests = Collections.singletonList(new AnsibleTest());
        this.nodeLoginTest = new NodeLoginTest();
    }

    public GuiLogicTestConfig(@Nullable /*@JsonProperty("amConnectionProxy")*/ be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy,
                              @Nullable /*@JsonProperty("amVersion")        */ String amVersion,
                              @Nullable /*@JsonProperty("sshKeys")          */ SSHKeys sshKeys,
                              @Nullable /*@JsonProperty("resources")        */ List<Resource> resources,
                              @Nullable /*@JsonProperty("ansibleTests")     */ List<AnsibleTest> ansibleTests,
                              @Nullable /*@JsonProperty("nodeLoginTest")    */ NodeLoginTest nodeLoginTest) {
        this.amConnectionProxy = amConnectionProxy == null ? new GuiLogicTestConfig().getAmConnectionProxy() : amConnectionProxy;
        this.amVersion = amVersion == null ? new GuiLogicTestConfig().getAmVersion() : amVersion;
        this.sshKeys = sshKeys == null ? new GuiLogicTestConfig().getSshKeys() : sshKeys;
        this.resources = resources == null ? new GuiLogicTestConfig().getResources() : resources;
        this.ansibleTests = ansibleTests == null ? new GuiLogicTestConfig().getAnsibleTests() : ansibleTests;
        this.nodeLoginTest = nodeLoginTest == null ? new GuiLogicTestConfig().getNodeLoginTest() : nodeLoginTest;
    }

    public GuiLogicTestConfig(@Nullable /*@JsonProperty("amConnectionProxy")*/ be.iminds.ilabt.jfed.testing.shared.Proxy amConnectionProxy,
                              @Nullable /*@JsonProperty("amVersion")        */ String amVersion,
                              @Nullable /*@JsonProperty("sshKeys")          */ SSHKeys sshKeys,
                              @Nullable /*@JsonProperty("resources")        */ List<Resource> resources,
                              @Nullable /*@JsonProperty("ansibleTests")     */ AnsibleTest ansibleTest,
                              @Nullable /*@JsonProperty("nodeLoginTest")    */ NodeLoginTest nodeLoginTest) {
        this(amConnectionProxy, amVersion, sshKeys, resources,
                ansibleTest == null ?
                        new GuiLogicTestConfig().getAnsibleTests() :
                        Collections.singletonList(ansibleTest),
                nodeLoginTest);}

    @JsonCreator
    public static GuiLogicTestConfig createFromJsonObject(@Nonnull Map<String,Object> props) {
        be.iminds.ilabt.jfed.testing.shared.Proxy proxy = MAPPER.convertValue(props.get("amConnectionProxy"), be.iminds.ilabt.jfed.testing.shared.Proxy.class);
        String amVersion = (String) props.get("amVersion");
        SSHKeys sshKeys = MAPPER.convertValue(props.get("sshKeys"), SSHKeys.class);
        List<Resource> resources = MAPPER.convertValue(props.get("resources"), new TypeReference<List<Resource>>() { });;
        List<AnsibleTest> ansibleTests;
        NodeLoginTest nodeLoginTest = MAPPER.convertValue(props.get("nodeLoginTest"), NodeLoginTest.class);

        if (props.containsKey("ansibleTest") || props.containsKey("ansibleTests")) {
            String key = props.containsKey("ansibleTests") ? "ansibleTests" : "ansibleTest";
            Object ansibleTestsObj = props.get(key);
            if (ansibleTestsObj instanceof List) {
                ansibleTests = MAPPER.convertValue(ansibleTestsObj, new TypeReference<List<AnsibleTest>>() { });
            } else {
                ansibleTests = Collections.singletonList(
                        MAPPER.convertValue(ansibleTestsObj, AnsibleTest.class)
                );
            }
        } else {
            ansibleTests = null;
        }

        return new GuiLogicTestConfig(proxy, amVersion, sshKeys, resources, ansibleTests, nodeLoginTest);
    }

    @Override
    @JsonIgnore
    public boolean isValid() {
        boolean res = getNodeLoginTest().isValid();
        res = res && getAmConnectionProxy().isValid();
        res = res && getSshKeys().isValid();
        for (Resource resource : getResources()) {
            res = res && resource.isValid();
        }
        for (AnsibleTest at : getAnsibleTests()) {
            res = res && at.isValid();
        }
        boolean amVersionValid = getAmVersion().equals("2") || getAmVersion().equals("3") || getAmVersion().equals("auto");
        if (!amVersionValid)
            throw new IllegalArgumentException("Config Error: Unsupported amVersion: \""+getAmVersion()+"\"");
//        res = res && amVersionValid;
        return res;
    }

    @JsonProperty
    @Nonnull
    public be.iminds.ilabt.jfed.testing.shared.Proxy getAmConnectionProxy() {
        return amConnectionProxy;
    }

    @JsonProperty
    @Nonnull
    public String getAmVersion() {
        return amVersion;
    }

    @JsonProperty
    @Nonnull
    public SSHKeys getSshKeys() {
        return sshKeys;
    }

    @JsonProperty
    @Nonnull
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    public List<Resource> getResources() {
        return resources;
    }

    @JsonProperty
    @Nonnull
    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    public List<AnsibleTest> getAnsibleTests() {
        return ansibleTests;
    }

    @JsonProperty
    @Nonnull
    public NodeLoginTest getNodeLoginTest() {
        return nodeLoginTest;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof GuiLogicTestConfig)) return false;

        GuiLogicTestConfig that = (GuiLogicTestConfig) o;

        if (!amConnectionProxy.equals(that.amConnectionProxy)) return false;
        if (!amVersion.equals(that.amVersion)) return false;
        if (!sshKeys.equals(that.sshKeys)) return false;
        if (!resources.equals(that.resources)) return false;
        if (!ansibleTests.equals(that.ansibleTests)) return false;
        return nodeLoginTest.equals(that.nodeLoginTest);
    }

    @Override
    public int hashCode() {
        int result = amConnectionProxy.hashCode();
        result = 31 * result + amVersion.hashCode();
        result = 31 * result + sshKeys.hashCode();
        result = 31 * result + resources.hashCode();
        result = 31 * result + ansibleTests.hashCode();
        result = 31 * result + nodeLoginTest.hashCode();
        return result;
    }

    @Nonnull
    @Override
    @JsonIgnore
    public String stringValue() {
        try {
            return MAPPER.writeValueAsString(this);
        } catch (JsonProcessingException e) {
            throw new RuntimeException("Failed to covert "+getClass().getName()+" to String", e);
        }
    }

    @Override
    public String toString() {
        return stringValue();
    }

    //What sort of data is captured/extracted?
    public enum ExtractDataType {
        TEXT, CSV, JSON
    }
    //What is the captured/extracted data about?
    public enum ExtractScope {
        TESTINSTANCE, SERVER, NODE
    }

    public static class ExtractSpec {
        //What to extract
        @Nonnull private final String delim;
        @Nullable private final String startDelim;
        @Nullable private final String endDelim;
        @Nullable private final String regex;    //a regex with a group, that is used to extract. Used instead of delim if present.
        @Nonnull private final ExtractDataType type;

        //Where to store it
        @Nonnull private final String name;
        @Nonnull private final ExtractScope scope;

        public ExtractSpec() {
            //startDelim and endDelim take priority over delim, if present
            //regex takes priority over all if present!
            this.delim = "$$$";
            this.startDelim = null;
            this.endDelim = null;
            this.regex = null;
            this.type = ExtractDataType.TEXT;
            this.name = "data";
            this.scope = ExtractScope.TESTINSTANCE;
        }

        @JsonCreator
        public ExtractSpec(@Nullable @JsonProperty("delim") String delim,
                           @Nullable @JsonProperty("startDelim") String startDelim,
                           @Nullable @JsonProperty("endDelim") String endDelim,
                           @Nullable @JsonProperty("regex") String regex,
                           @Nullable @JsonProperty("type") ExtractDataType type,
                           @Nullable @JsonProperty("name") String name,
                           @Nullable @JsonProperty("scope") ExtractScope scope) {
            this.delim = delim != null ? delim : new ExtractSpec().getDelim();
            this.startDelim = startDelim == null || startDelim.trim().isEmpty() ? null : startDelim;
            this.endDelim = endDelim == null || endDelim.trim().isEmpty() ? null : endDelim;
            this.regex = regex == null || regex.trim().isEmpty() ? null : regex;
            this.type = type != null ? type : new ExtractSpec().getType();
            this.name = name != null ? name : new ExtractSpec().getName();
            this.scope = scope != null ? scope : new ExtractSpec().getScope();
        }

        @JsonProperty
        @Nonnull
        public String getDelim() {
            return delim;
        }

        @JsonProperty
        @Nullable
        public String getStartDelim() {
            return startDelim;
        }

        @JsonProperty
        @Nullable
        public String getEndDelim() {
            return endDelim;
        }

        @JsonProperty
        @Nullable
        public String getRegex() {
            return regex;
        }

        @JsonIgnore
        @Nonnull
        public String getStartDelimOrFallback() {
            return startDelim == null ? delim : startDelim;
        }

        @JsonIgnore
        @Nonnull
        public String getEndDelimOrFallback() {
            return endDelim == null ? delim : endDelim;
        }

        @JsonProperty
        @Nonnull
        public ExtractDataType getType() {
            return type;
        }

        @JsonProperty
        @Nonnull
        public String getName() {
            return name;
        }

        @JsonProperty
        @Nonnull
        public ExtractScope getScope() {
            return scope;
        }

        @JsonIgnore
        public boolean isValid() {
            //delim is fallback, so may never be null
            if (delim.trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.extract.delim may not be empty");
            }
            if (name.trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.extract.name may not be empty");
            }
            return true;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context", "playbookExe"})
    public static class AnsibleTest {
        @Nonnull private final Boolean enabled;
        @Nonnull private final Boolean debug;
        @Nullable private final String playbookUrl; //may be yml file, or archive (zip, tar, ...)
        @Nullable private final String playbookFileNameInArchive; //if playbook URL points to an archive, this is the name of the ansible playbook in it
        @Nullable private final String playbookContent;
        @Nullable private final String failureRegex;
        @Nullable private final String warningRegex;
        @Nullable private final String successRegex;
        @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy proxy;
        @Nonnull private final List<ExtractSpec> extract;
        private final int timeoutInSec; //<= 0 means no timeout

        public AnsibleTest() {
            this.enabled = false;
            this.proxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
            this.debug = false;
            this.playbookUrl = null;
            this.playbookFileNameInArchive = null;
            this.playbookContent = null;
            this.failureRegex = "(failed=[^0]|unreachable=[^0])"; //"failed=[^0]";
            this.warningRegex = null;//"unreachable=[^0]";
            this.successRegex = "ok=[^0]";
            this.extract = Collections.emptyList();
            this.timeoutInSec = 20*60;
        }

        @JsonCreator
        public AnsibleTest(@Nullable @JsonProperty("enabled") Boolean enabled,
                           @Nullable @JsonProperty("proxy") be.iminds.ilabt.jfed.testing.shared.Proxy proxy,
                           @Nullable @JsonProperty("debug") Boolean debug,
                           @Nullable @JsonProperty("playbookUrl") String playbookUrl,
                           @Nullable @JsonProperty("playbookFileNameInArchive") String playbookFileNameInArchive,
                           @Nullable @JsonProperty("playbookContent") String playbookContent,
                           @Nullable @JsonProperty("failureRegex") String failureRegex,
                           @Nullable @JsonProperty("warningRegex") String warningRegex,
                           @Nullable @JsonProperty("successRegex") String successRegex,
                           @Nullable @JsonProperty("extract") List<ExtractSpec> extract,
                           @Nullable @JsonProperty("timeoutInSec") Integer timeoutInSec) {
            this.enabled = enabled == null ? new AnsibleTest().isEnabled() : enabled;
            this.proxy = proxy == null ? new AnsibleTest().getProxy() : proxy;
            this.debug = debug == null ? new AnsibleTest().getDebug() : debug;
            this.playbookUrl = trimAndToNullIfEmpty(playbookUrl);
            this.playbookFileNameInArchive = trimAndToNullIfEmpty(playbookFileNameInArchive);
            this.playbookContent = toNullIfEmpty(playbookContent);
            this.failureRegex = toNullableDefaultIfEmpty(failureRegex, () -> new AnsibleTest().getFailureRegex());
            this.warningRegex = toNullableDefaultIfEmpty(warningRegex, () -> new AnsibleTest().getWarningRegex());
            this.successRegex = toNullableDefaultIfEmpty(successRegex, () -> new AnsibleTest().getSuccessRegex());
            this.extract = extract != null ?
                    Collections.unmodifiableList(new ArrayList<>(extract)) :
                    new AnsibleTest().getExtract();
            this.timeoutInSec = timeoutInSec == null ? 20*60 : timeoutInSec;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!isEnabled())
                return true;

            if (getPlaybookContent() == null && getPlaybookUrl() == null) {
                throw new IllegalArgumentException("Config Error: You need to specify either ansible.playbookContent or ansible.playbookUrl");
            }
            if (getPlaybookContent() != null && getPlaybookUrl() != null) {
                throw new IllegalArgumentException("Config Error: You need to specify either ansible.playbookContent or ansible.playbookUrl, but not both");
            }
            if (getPlaybookContent() != null && getPlaybookContent().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.playbookContent may not be empty when specified");
            }
            if (getPlaybookUrl() != null) {
                if (getPlaybookUrl().trim().isEmpty()) {
                    throw new IllegalArgumentException("Config Error: ansible.playbookUrl may not be empty when specified");
                }
                if (!(getPlaybookUrl().trim().startsWith("http://") || getPlaybookUrl().trim().startsWith("https://"))) {
                    if (getPlaybookUrl().trim().startsWith("git ") || getPlaybookUrl().trim().startsWith("github ")) {
                        //ok, special case: git/github with subdir and branch
                        //subdir may be / for no subdir

                        //syntax: git[hub] <https://<githubTLD>/<org>/<repo>.git> [<subdir> [<branch>]]
                        if (playbookFileNameInArchive == null) {
                            throw new IllegalArgumentException("Config Error: ansible.playbookUrl refers to git (\"" + getPlaybookUrl() + "\"), but playbookFileNameInArchive is not set");
                        }
                    } else {
                        throw new IllegalArgumentException("Config Error: ansible.playbookUrl does not appear to be a URL: \"" + getPlaybookUrl() + "\"");
                    }
                } else {
                    if (isProbablyArchiveUrl(getPlaybookUrl().trim()) && playbookFileNameInArchive == null) {
                        throw new IllegalArgumentException("Config Error: ansible.playbookUrl appear to be an archive(\"" + getPlaybookUrl() + "\"), but playbookFileNameInArchive is not set");
                    }
                }
            }
            if (getPlaybookUrl() == null && playbookFileNameInArchive != null) {
                throw new IllegalArgumentException("Config Error: ansible.playbookFileNameInArchive only makes sense if ansible.playbookUrl is used (and points to an archive)");
            }
            if (getFailureRegex() != null && getFailureRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.failureRegex may not be empty");
            }
            if (getWarningRegex() != null && getWarningRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.warningRegex may not be empty");
            }
            if (getSuccessRegex() != null && getSuccessRegex().trim().isEmpty()) {
                throw new IllegalArgumentException("Config Error: ansible.successRegex may not be empty");
            }

            for (ExtractSpec extractSpec : getExtract()) {
                boolean extractValid = extractSpec.isValid();
                if (!extractValid)
                    return false;
            }

            return proxy.isValid();
        }

        private static boolean isProbablyArchiveUrl(String url) {
            if (url.endsWith(".zip"))
                return true;
            if (url.endsWith(".tar"))
                return true;
            if (url.endsWith(".tar.gz"))
                return true;
            if (url.endsWith(".tar.bz2"))
                return true;
            if (url.endsWith(".rar"))
                return true;
            return false;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Timeout of the test in seconds. <= 0 means to timeout. Default is 20 minutes.")
        public int getTimeoutInSec() {
            return timeoutInSec;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="enables the runAnsible() step of the test")
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="(SSH) proxy to use for the ansible tests")
        @Nonnull
        public be.iminds.ilabt.jfed.testing.shared.Proxy getProxy() {
            return proxy;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="sets the debug flag when calling ansible")
        public boolean getDebug() {
            return debug;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="URL of the PlayBook to execute (either playbookUrl or playbookUrl must be provided, and not both)")
        @Nullable
        public String getPlaybookUrl() {
            return playbookUrl;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Name of the PlayBook inside the archive pointed to by playbookUrl")
        @Nullable
        public String getPlaybookFileNameInArchive() {
            return playbookFileNameInArchive;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Content of the PlayBook to execute (either playbookUrl or playbookUrl must be provided, and not both)")
        @Nullable
        public String getPlaybookContent() {
            return playbookContent;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered failed.")
        @Nullable
        public String getFailureRegex() {
            return failureRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered warning. (Unless it is failed.)")
        @Nullable
        public String getWarningRegex() {
            return warningRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="When this regex matches a line in the ansible output, the test state is considered succes. (Unless it is failed or warning.)")
        @Nullable
        public String getSuccessRegex() {
            return successRegex;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Data to extract from ansible result")
        @Nonnull
        public List<ExtractSpec> getExtract() {
            return extract;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof AnsibleTest)) return false;

            AnsibleTest that = (AnsibleTest) o;

            if (!enabled.equals(that.enabled)) return false;
            if (!debug.equals(that.debug)) return false;
            if (playbookUrl != null ? !playbookUrl.equals(that.playbookUrl) : that.playbookUrl != null) return false;
            if (playbookFileNameInArchive != null ? !playbookFileNameInArchive.equals(that.playbookFileNameInArchive) : that.playbookFileNameInArchive != null)
                return false;
            if (playbookContent != null ? !playbookContent.equals(that.playbookContent) : that.playbookContent != null)
                return false;
            if (failureRegex != null ? !failureRegex.equals(that.failureRegex) : that.failureRegex != null)
                return false;
            if (warningRegex != null ? !warningRegex.equals(that.warningRegex) : that.warningRegex != null)
                return false;
            if (successRegex != null ? !successRegex.equals(that.successRegex) : that.successRegex != null)
                return false;
            if (!proxy.equals(that.proxy)) return false;
            return extract.equals(that.extract);
        }

        @Override
        public int hashCode() {
            int result = enabled.hashCode();
            result = 31 * result + debug.hashCode();
            result = 31 * result + (playbookUrl != null ? playbookUrl.hashCode() : 0);
            result = 31 * result + (playbookFileNameInArchive != null ? playbookFileNameInArchive.hashCode() : 0);
            result = 31 * result + (playbookContent != null ? playbookContent.hashCode() : 0);
            result = 31 * result + (failureRegex != null ? failureRegex.hashCode() : 0);
            result = 31 * result + (warningRegex != null ? warningRegex.hashCode() : 0);
            result = 31 * result + (successRegex != null ? successRegex.hashCode() : 0);
            result = 31 * result + proxy.hashCode();
            result = 31 * result + extract.hashCode();
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class NodeLoginTest {
        private final boolean enabled;
        private final long deadlineSeconds;
        @Nullable private final String preferredUser;
        @Nullable private final Integer forceIpVersion;
        private final boolean useExternalSsh;
        @Nonnull private final be.iminds.ilabt.jfed.testing.shared.Proxy proxy;

        public NodeLoginTest() {
            this.enabled = true;
            this.deadlineSeconds = 20 * 60;
            this.preferredUser = null;
            this.forceIpVersion = null;
            this.useExternalSsh = false;
            this.proxy = new be.iminds.ilabt.jfed.testing.shared.Proxy();
        }

        @JsonCreator
        public NodeLoginTest(@Nullable @JsonProperty("enabled") Boolean enabled,
                             @Nullable @JsonProperty("deadlineSeconds") Long deadlineSeconds,
                             @Nullable @JsonProperty("preferredUser") String preferredUser,
                             @Nullable @JsonProperty("forceIpVersion") Integer forceIpVersion,
                             @Nullable @JsonProperty("useExternalSsh") Boolean useExternalSsh,
                             @Nullable @JsonProperty("proxy") be.iminds.ilabt.jfed.testing.shared.Proxy proxy) {
            this.enabled = enabled == null ? new NodeLoginTest().isEnabled() : enabled;
            this.deadlineSeconds = deadlineSeconds == null ? new NodeLoginTest().getDeadlineSeconds() : deadlineSeconds;
            this.preferredUser = preferredUser;
            this.forceIpVersion = forceIpVersion;
            this.useExternalSsh = useExternalSsh == null ? new NodeLoginTest().getUseExternalSsh() : useExternalSsh;
            this.proxy = proxy == null ? new NodeLoginTest().getProxy() : proxy;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!isEnabled())
                return true;

            if (deadlineSeconds <= 0)
                throw new IllegalArgumentException("Config Error: " +
                        "nodelogin.deadlineSeconds <= 0 makes no sense: "+deadlineSeconds);

            boolean ipVersionValid = getForceIpVersion() == null ||
                    getForceIpVersion().equals(4) ||
                    getForceIpVersion().equals(6);
            if (!ipVersionValid)
                throw new IllegalArgumentException("Config Error: " +
                        "Unsupported forceIpVersion: \""+getForceIpVersion()+"\"");

            return proxy.isValid();
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        public long getDeadlineSeconds() {
            return deadlineSeconds;
        }

        @JsonProperty
        @Nullable
        public String getPreferredUser() {
            return preferredUser;
        }

        @JsonProperty
        @Nullable
        public Integer getForceIpVersion() {
            return forceIpVersion;
        }

        @JsonProperty
        public boolean getUseExternalSsh() {
            return useExternalSsh;
        }

        @JsonProperty
        @Nonnull
        public be.iminds.ilabt.jfed.testing.shared.Proxy getProxy() {
            return proxy;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof NodeLoginTest)) return false;

            NodeLoginTest that = (NodeLoginTest) o;

            if (enabled != that.enabled) return false;
            if (deadlineSeconds != that.deadlineSeconds) return false;
            if (useExternalSsh != that.useExternalSsh) return false;
            if (preferredUser != null ? !preferredUser.equals(that.preferredUser) : that.preferredUser != null)
                return false;
            if (forceIpVersion != null ? !forceIpVersion.equals(that.forceIpVersion) : that.forceIpVersion != null)
                return false;
            return proxy.equals(that.proxy);
        }

        @Override
        public int hashCode() {
            int result = (enabled ? 1 : 0);
            result = 31 * result + (int) (deadlineSeconds ^ (deadlineSeconds >>> 32));
            result = 31 * result + (preferredUser != null ? preferredUser.hashCode() : 0);
            result = 31 * result + (forceIpVersion != null ? forceIpVersion.hashCode() : 0);
            result = 31 * result + (useExternalSsh ? 1 : 0);
            result = 31 * result + proxy.hashCode();
            return result;
        }
    }

    public enum SSHKeySource { TEST_USER, RANDOM };

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class SSHKeys {
        @Nonnull private final SSHKeySource source;

        //maybe future, but not sure how to do this securily from fedmon:
        // , fixedContent, fixedFile
//        @Nonnull private final String publicKeyContent;
//        @Nonnull private final String privateKeyContent;
//        @Nonnull private final String publicKeyFile;
//        @Nonnull private final String privateKeyFile;
//        @Nonnull private final String privateKeyPassword;


        public SSHKeys() {
            this.source = SSHKeySource.RANDOM;
        }

        @JsonCreator
        public SSHKeys(@Nullable @JsonProperty("source") SSHKeySource source) {
            this.source = source == null ? new SSHKeys().getSource() : source;
        }

        @JsonIgnore
        public boolean isValid() {
            return true;
        }

        @JsonProperty
        @Nonnull
        public SSHKeySource getSource() {
            return source;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof SSHKeys)) return false;

            SSHKeys sshKeys = (SSHKeys) o;

            return source == sshKeys.source;
        }

        @Override
        public int hashCode() {
            return source.hashCode();
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Cleanup {
        private final boolean enabled;
        private final boolean ignoreErrors;
        private final long maxWaitTimeMinutes;

        public Cleanup() {
            this.enabled = true;
            this.ignoreErrors = false;
            this.maxWaitTimeMinutes = 5;
        }

        @JsonCreator
        public Cleanup(@Nullable @JsonProperty("enabled") Boolean enabled,
                       @Nullable @JsonProperty("ignoreErrors") Boolean ignoreErrors,
                       @Nullable @JsonProperty("maxWaitTimeMinutes") Long maxWaitTimeMinutes) {
            this.enabled = enabled == null ? new Cleanup().isEnabled() : enabled;
            this.ignoreErrors = ignoreErrors == null ? new Cleanup().getIgnoreErrors() : ignoreErrors;
            this.maxWaitTimeMinutes = maxWaitTimeMinutes == null ? new Cleanup().getMaxWaitTimeMinutes() : maxWaitTimeMinutes;
        }

        @JsonIgnore
        public boolean isValid() {
            if (!enabled)
                return true;

            if (maxWaitTimeMinutes < 0)
                throw new IllegalArgumentException("Config Error: cleanup.maxWaitTimeMinutes < 0 makes no sense");

            return true;
        }

        @JsonProperty
        public boolean isEnabled() {
            return enabled;
        }

        @JsonProperty
        public boolean getIgnoreErrors() {
            return ignoreErrors;
        }

        @JsonProperty
        public long getMaxWaitTimeMinutes() {
            return maxWaitTimeMinutes;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Cleanup)) return false;

            Cleanup cleanup = (Cleanup) o;

            if (enabled != cleanup.enabled) return false;
            if (ignoreErrors != cleanup.ignoreErrors) return false;
            return maxWaitTimeMinutes == cleanup.maxWaitTimeMinutes;
        }

        @Override
        public int hashCode() {
            int result = (enabled ? 1 : 0);
            result = 31 * result + (ignoreErrors ? 1 : 0);
            result = 31 * result + (int) (maxWaitTimeMinutes ^ (maxWaitTimeMinutes >>> 32));
            return result;
        }
    }

    @JsonInclude(JsonInclude.Include.NON_DEFAULT)
    @JsonIgnoreProperties(value={"@context"})
    public static class Resource extends ExperimentSetupConfig {
        @Nonnull private final Cleanup cleanup;

        private final boolean skipFinalManifestCheck;
        private final boolean requireNodesInManifest;

        public Resource() {
            super();
            this.cleanup = new Cleanup();
            this.skipFinalManifestCheck = false;
            this.requireNodesInManifest = true;
        }
        @JsonCreator
        public Resource(@Nullable @JsonProperty("slice") Slice slice,
                        @Nullable @JsonProperty("requestRSpec") RequestRSpec requestRSpec,
                        @Nullable @JsonProperty("eSpec") ESpec eSpec,
                        @Nullable @JsonProperty("provision") Provision provision,
                        @Nullable @JsonProperty("waitForReady") WaitForReady waitForReady,
                        @Nullable @JsonProperty("cleanup") Cleanup cleanup,
                        @Nullable @JsonProperty("overrideESpecRSpec") Boolean overrideESpecRSpec,
                        @Nullable @JsonProperty("runLinkTest") Boolean runLinkTest,
                        @Nullable @JsonProperty("skipFinalManifestCheck") Boolean skipFinalManifestCheck,
                        @Nullable @JsonProperty("requireNodesInManifest") Boolean requireNodesInManifest) {
            super(slice, requestRSpec, eSpec, provision, waitForReady, overrideESpecRSpec, runLinkTest);
            this.cleanup = cleanup == null ? new Resource().getCleanup() : cleanup;
            this.skipFinalManifestCheck = skipFinalManifestCheck == null ? new Resource().getSkipFinalManifestCheck() : skipFinalManifestCheck;
            this.requireNodesInManifest = requireNodesInManifest == null ? new Resource().getRequireNodesInManifest() : requireNodesInManifest;
        }

        @JsonIgnore
        public boolean isValid() {
            boolean res = super.isValid();
            res = res && getCleanup().isValid();
            return res;
        }

        @JsonProperty
        @ApiTestConfigProperty(description="Skip the detailed manifest test in the \"checkManifest\" step? If skipped, very basic testing of the manifest is still done.")
        public boolean getSkipFinalManifestCheck() {
            return skipFinalManifestCheck;
        }


        @JsonProperty
        public boolean getRequireNodesInManifest() {
            return requireNodesInManifest;
        }

        @JsonProperty
        @Nullable
        @Override
        @ApiTestConfigProperty(noEdit = true) //do not edit -> type is too complex
        public ExperimentSetupConfig.RequestRSpec getRequestRSpec() {
            return requestRSpec;
        }


        @JsonProperty
        @Nonnull
        public Cleanup getCleanup() {
            return cleanup;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (!(o instanceof Resource)) return false;
            if (!super.equals(o)) return false;
            Resource resource = (Resource) o;
            return skipFinalManifestCheck == resource.skipFinalManifestCheck &&
                    requireNodesInManifest == resource.requireNodesInManifest &&
                    Objects.equals(cleanup, resource.cleanup);
        }

        @Override
        public int hashCode() {
            return Objects.hash(super.hashCode(), cleanup, skipFinalManifestCheck, requireNodesInManifest);
        }
    }




    public NodeLoginTestStepConfig generateNodeLoginTestStepConfig(@Nonnull GeniUserProvider geniUserProvider) {
        NodeLoginTestStepConfig.SSHKeySource sshKeySource;
        PublicKey publicKey;
        PrivateKey privateKey;
        switch (getSshKeys().getSource()) {
            case RANDOM: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.RANDOM;
                publicKey = null;
                privateKey = null;
                break;
            }
            case TEST_USER: {
                sshKeySource = NodeLoginTestStepConfig.SSHKeySource.PROVIDED;
                assert geniUserProvider.isUserLoggedIn();
                GeniUser user = geniUserProvider.getLoggedInGeniUser();
                assert user != null;
                publicKey = user.getPublicKey();
                privateKey = user.getPrivateKey();
                break;
            }
            default: throw new IllegalArgumentException("Unsupported ssh keys source: "+getSshKeys().getSource());
        }


        return new NodeLoginTestStepConfig(sshKeySource,
                publicKey,
                privateKey,
                getNodeLoginTest().getUseExternalSsh(),
                getNodeLoginTest().getProxy(),
                getNodeLoginTest().getDeadlineSeconds() * 1000l,
                getNodeLoginTest().getPreferredUser(),
                getNodeLoginTest().getForceIpVersion());
    }


    @Contract("null -> null")
    @Nullable private static String trimAndToNullIfEmpty(@Nullable String orig) {
        if (orig == null)
            return null;
        if (orig.trim().isEmpty())
            return null;
        return orig.trim();
    }

    @Contract("null -> null")
    @Nullable private static String toNullIfEmpty(@Nullable String orig) {
        if (orig == null)
            return null;
        if (orig.trim().isEmpty())
            return null;
        return orig.trim();
    }

    @FunctionalInterface
    public interface NonNullSupplier<T> {
        @Nonnull T get();
    }

    @Contract("null,null -> null; _,!null->!null")
    @Nullable private static String trimAndToDefaultIfEmpty(@Nullable String orig,
                                                            @Nullable NonNullSupplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig.trim();
    }

    @Contract("null,null -> null; _,!null->!null")
    @Nullable private static String toDefaultIfEmpty(@Nullable String orig,
                                                     @Nullable NonNullSupplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig;
    }

    @Contract("null,null -> null")
    @Nullable private static String toNullableDefaultIfEmpty(@Nullable String orig,
                                                             @Nullable Supplier<String> defaultString) {
        if (orig == null)
            return defaultString == null ? null : defaultString.get();
        if (orig.trim().isEmpty())
            return defaultString == null ? null : defaultString.get();
        return orig;
    }
}
