package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.tasks.RecoverSliceTaskInteraction;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

public class RecoverSliceTaskCliInteraction implements RecoverSliceTaskInteraction {
    private static final Logger LOG = LoggerFactory.getLogger(RecoverSliceTaskCliInteraction.class);

    @Nonnull private final Context context;
    private Experiment experiment;

    public RecoverSliceTaskCliInteraction(@Nonnull Context context) {
        this.context = context;
    }

    @Override
    public void onCouldNotRetrieve(@Nullable Throwable exception) {
        LOG.error("Could not retrieve slice", exception);
        context.getActionLogger().errorPrintln("Could not retrieve slice");
    }

    @Override
    public void onUnknownAuthoritiesDetected(@Nonnull String sliceName, @Nonnull List<GeniUrn> unknownAuthorities) throws InterruptedException {
        String auths = unknownAuthorities.stream().map(GeniUrn::toString).collect(Collectors.joining(", "));
        LOG.error("Unknown Authorities Detected: "+auths);
        context.getActionLogger().errorPrintln("Unknown Authorities Detected: "+auths);
    }

    @Override
    public boolean onNoActiveResources(@Nonnull String sliceName) throws InterruptedException {
        LOG.error("No active resources for "+sliceName);
        context.getActionLogger().errorPrintln("No active resources for "+sliceName);
        //TODO: support this using RSpec argument, if present
        return false;
    }

    @Nonnull
    @Override
    public Optional<Set<Server>> requestAuthorities() throws InterruptedException {
        //TODO: get from --rspec argument if present
        return Optional.empty();
    }

    @Override
    public void onCouldNotFindInfoWhenUserSelected(@Nonnull String sliceName) throws InterruptedException {
        LOG.error("Could not find active resources on the authorities in the provided manifest RSpec for "+sliceName);
        context.getActionLogger().errorPrintln("Could not find active resources on the authorities in the provided manifest RSpec for "+sliceName);
    }

    @Override
    public void onCouldNotFindInfoWhenAutoSelected(@Nonnull String sliceName, @Nonnull Set<Server> allInvolvedAuthorities) {
        String serverNames = allInvolvedAuthorities.stream().map(Server::getName).collect(Collectors.toList()).toString();
        LOG.error("Could not find information about "+sliceName+" on the involved testbeds: "+serverNames);
        context.getActionLogger().errorPrintln("Could not find information about "+sliceName+" on the involved testbeds: "+serverNames);
    }

    @Override
    public void onFinalFailed(@Nonnull String sliceName, @Nullable String message, @Nullable Throwable exception) {
        LOG.error("Recover failed for "+sliceName+": "+message, exception);
        context.getActionLogger().errorPrintln("Recover failed for "+sliceName+": "+message);
    }

    @Override
    public void onFinalSucces(@Nonnull Experiment experiment) {
        LOG.error("Successfully recovered experiment");
        this.experiment = experiment;
    }

    @Nullable
    public Experiment getExperiment() {
        return experiment;
    }
}
