package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import javafx.fxml.FXML;
import javafx.scene.control.ComboBox;
import javafx.scene.layout.HBox;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * User: twalcari
 * Date: 9/10/2014
 * Time: 09:44
 */
public class ExoGeniConfigTab extends AdaptivePropertiesTab<FXRspecNode> {
    private static final Logger LOG = LoggerFactory.getLogger(ExoGeniConfigTab.class);

    private final AdaptivePropertiesDialog adaptivePropertiesDialog;

    @FXML
    private ComboBox<String> resourceTypeComboBox;


    @FXML
    private HBox exoGeniHelp;
    @FXML
    private HBox genericHelp;


    private final BrowserUtil browserUtil;

    private static final String[] EXOGENI_VM_TYPES =
            new String[]{"default-vm", "xo.small", "xo.medium", "xo.large", "xo.xlarge"};
//            new String[]{"default-vm", "xo.small", "xo.medium", "xo.large", "xo.xlarge", "exogeni-m4"};

    public ExoGeniConfigTab(AdaptivePropertiesDialog adaptivePropertiesDialog,
                            BrowserUtil browserUtil) {
        super("Resource Type");
        this.adaptivePropertiesDialog = adaptivePropertiesDialog;
        this.browserUtil = browserUtil;

        setContent(FXMLUtil.createFromFXML(this));

//        resourceTypeComboBox.getItems().addAll(CORRECT_TYPES);
        adaptivePropertiesDialog.selectedResourceProperty().addListener(
                (prop, oldVal, newVal) -> {
                    FXRspecNode rspecNode = adaptivePropertiesDialog.getRspecNode();
                    if (newVal != null && newVal.getAllowedSliverTypes() != null &&
                            rspecNode.getSliverTypeName() != null &&
                            newVal.getSliverType() != null &&
                            !newVal.getAllowedSliverTypes().contains(rspecNode.getSliverTypeName())) {
                        rspecNode.setSliverTypeName(newVal.getSliverType());
                    }
                    load(rspecNode);
                }
        );
    }

    @Override
    public void load(FXRspecNode rspecNode) {
        boolean isEG = isExoGeni(rspecNode);

        exoGeniHelp.setVisible(isEG);
        genericHelp.setVisible(!isEG);

        setText(isEG ? "ExoGeni Resource Type" : "Resource Type");

        Resource resource = adaptivePropertiesDialog.getSelectedResource();
        if (resource == null) {
            LOG.warn("adaptivePropertiesDialog.getSelectedResource() is null");
            resourceTypeComboBox.getItems().clear();
            resourceTypeComboBox.getSelectionModel().clearSelection();
            return;
        }

        Set<String> allowedTypes = new HashSet<>();
        if (resource.getAllowedSliverTypes() != null)
            allowedTypes.addAll(resource.getAllowedSliverTypes());
        if (isEG)
            allowedTypes.addAll(Arrays.asList(EXOGENI_VM_TYPES));

        LOG.debug("Setting allowedTypes (isEG={}): {}", isEG, allowedTypes);
        resourceTypeComboBox.getItems().setAll(allowedTypes);

        //initialize tables with information from node
        String sliverTypeName = rspecNode.getSliverTypeName();
        if (sliverTypeName == null && allowedTypes.contains("default-vm"))
            sliverTypeName = "default-vm";;
        if (sliverTypeName == null && !allowedTypes.isEmpty())
            sliverTypeName = allowedTypes.iterator().next();
        if (sliverTypeName != null) {
            if (!allowedTypes.contains(sliverTypeName)) {
                //also allow any type the user specified manually in the RSpec
                resourceTypeComboBox.getItems().add(sliverTypeName);
                LOG.warn("sliver_type {} unknown, will still show in list (known={})",
                        sliverTypeName, allowedTypes);
            }
            resourceTypeComboBox.getSelectionModel().select(sliverTypeName);
        }
    }

    @Override
    public void save(FXRspecNode rspecNode) {
        String selectedTypeName = resourceTypeComboBox.getSelectionModel().getSelectedItem();
        if (selectedTypeName != null)
            rspecNode.setSliverTypeName(selectedTypeName);
        LOG.debug("After setting sliver type name, slivertype it is: " + rspecNode.getSliverType());
    }

    @Override
    public void onTabShown() {

    }

    public boolean isExoGeni(FXRspecNode rspecNode) {
        Resource resource = adaptivePropertiesDialog.getSelectedResource();
        if (resource == null) return false;
        Server server = resource.getServer();
        if (server == null) return false;
        return server.isServerType("exogeni") || server.isServerType("orca");
    }

    @Override
    public boolean shouldBeVisible(FXRspecNode rspecNode) {
        if (isExoGeni(rspecNode))
            return true;

        Resource resource = adaptivePropertiesDialog.getSelectedResource();
        if (resource == null) return false;
        assert resource.getAllowedSliverTypes() != null;
        if (resource.getAllowedSliverTypes() == null) return false;
        LOG.debug("resource {} has allowedSliverTypes = {}", resource.getId(), resource.getAllowedSliverTypes());
        return resource.getAllowedSliverTypes().stream()
                .filter((st) -> !st.equalsIgnoreCase("default-vm"))
                .count() > 1;
    }

    @Override
    public String validateConfiguration() {
        return null;
    }


    public void onExoResourceHelpLinkClicked() {
        try {
            browserUtil.openUrlInBrowser(
                    new URI("https://wiki.exogeni.net/doku.php?id=public:experimenters:resource_types:start")
            );
        } catch (URISyntaxException e) {
            LOG.error("Should not occur", e);
        }
    }
}
